#define com_maroontress_libcgi_mime_ContentType_IMPORT
#define com_maroontress_libcgi_request_FileSegment_IMPORT
#define com_maroontress_libcgi_request_MimePart_IMPORT
#define com_maroontress_libcgi_util_Tub_IMPORT
#define std_io_OutputTube_IMPORT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <com/maroontress/libcgi/mime/ContentType.h>
#include <com/maroontress/libcgi/request/FileSegment.h>
#include <com/maroontress/libcgi/request/MimePart.h>
#include <com/maroontress/libcgi/util/Tub.h>
#include <std/io/OutputTube.h>

struct MimePart {
    FILE *fp;
    struct FileSegment segment;
    struct FileSegment bodySegment;
    struct ContentType *contentType;
    struct ContentType *disposition;
    char *cache;
};

static int
isSpace(int c)
{
    return (c == ' ' || c == '\t');
}

static int
nextByte(FILE *fp)
{
    int c;

    c = getc(fp);
    return (fseek(fp, -1, SEEK_CUR) < 0) ? EOF : c;
}

static int
getLine(FILE *fp, struct Tub *t)
{
    int c;

    Tub_clear(t);
    do {
	while ((c = getc(fp)) != EOF && Tub_writeByte(t, c) == 0 && c != '\n')
	    ;
	if (c != '\n' || (c = nextByte(fp)) == EOF)
	    return -1;
    } while (isSpace(c));
    return Tub_writeByte(t, 0);
}

static void
unfold(char *d)
{
    char *s = d;

    while (*s) {
        if (*s == '\n') {
            *d = ' ';
            for (++s; isSpace(*s); ++s)
                ;
        }
	else {
	    *d = *s;
	    ++s;
	}
	++d;
    }
    *d = 0;
}

struct CString {
    const char *str;
    size_t len;
};
#define CString_new(x) {x, sizeof(x) - 1}
static const struct CString CONTENT_TYPE =
    CString_new("Content-Type:");
static const struct CString CONTENT_DISPOSITION =
    CString_new("Content-Disposition:");

static int
setField(struct ContentType **field, const struct CString *name, const char *p)
{
    if (strncasecmp(p, name->str, name->len) != 0)
	return 1;
    for (p += name->len; *p && isSpace(*p); ++p)
	;
    ContentType_free(*field);
    if ((*field = ContentType_create(p)) == NULL)
	return -1;
    return 0;
}

static int
parseField(struct MimePart *mp, char *str)
{
    int s;

    if (strcmp(str, "\r\n") == 0) {
	return 1;
    }
    unfold(str);
    if ((s = setField(&mp->contentType, &CONTENT_TYPE, str)) <= 0
	|| (s = setField(&mp->disposition, &CONTENT_DISPOSITION, str)) <= 0) {
	return s;
    }
    /* Τʤեɤ̵뤹 */
    return 0;
}

static int
initializeWithTub(struct MimePart *mp, struct Tub *tub)
{
    char *str;
    int s;

    if (fseeko(mp->fp, mp->segment.offset, SEEK_SET) < 0)
	return -1;
    while ((s = getLine(mp->fp, tub)) == 0) {
	if ((str = Tub_createByteArray(tub)) == NULL)
	    return -1;
	s = parseField(mp, str);
	free(str);
	if (s < 0)
	    return -1;
	if (s == 1)
	    break;
    }
    if (s < 0)
	return -1;

    if ((mp->bodySegment.offset = ftello(mp->fp)) < 0
	|| (mp->bodySegment.length = (mp->segment.offset + mp->segment.length
				      - mp->bodySegment.offset)) < 0) {
	return -1;
    }
    return 0;
}

static int
initialize(struct MimePart *mp)
{
    struct Tub *tub;
    int s;

    if ((tub = Tub_create()) == NULL)
	return -1;
    s = initializeWithTub(mp, tub);
    Tub_free(tub);
    return s;
}

void
MimePart_free(struct MimePart *mp)
{
    if (mp == NULL)
	return;
    ContentType_free(mp->contentType);
    ContentType_free(mp->disposition);
    free(mp->cache);
    free(mp);
}

struct MimePart *
MimePart_create(FILE *fp, const struct FileSegment *segment)
{
    struct MimePart *mp;

    if ((mp = (struct MimePart *)malloc(sizeof(*mp))) == NULL)
        return NULL;
    mp->fp = fp;
    mp->contentType = NULL;
    mp->disposition = NULL;
    mp->cache = NULL;
    mp->segment = *segment;

    if (initialize(mp) < 0) {
	MimePart_free(mp);
	return NULL;
    }
    return mp;
}

struct ContentType *
MimePart_getContentType(struct MimePart *mp)
{
    return mp->contentType;
}

struct ContentType *
MimePart_getDisposition(struct MimePart *mp)
{
    return mp->disposition;
}

char *
MimePart_getContent(struct MimePart *mp)
{
    unsigned char *data;

    if (mp->cache != NULL)
	return mp->cache;
    if ((data = (unsigned char *)malloc(mp->bodySegment.length + 1)) == NULL)
	return NULL;
    if (mp->bodySegment.length > 0
	&& (fseeko(mp->fp, mp->bodySegment.offset, SEEK_SET) < 0
	    || fread(data, mp->bodySegment.length, 1, mp->fp) < 1)) {
	free(data);
	return NULL;
    }
    data[mp->bodySegment.length] = 0;
    mp->cache = data;
    return data;
}

off_t
MimePart_write(struct MimePart *mp, union OutputTube out)
{
    off_t fileSize;
    unsigned char data[4096];
    size_t size;

    if (fseeko(mp->fp, mp->bodySegment.offset, SEEK_SET) < 0)
	return -1;
    fileSize = mp->bodySegment.length;
    while (fileSize > 0) {
	size = (fileSize > sizeof(data)) ? sizeof(data) : (size_t)fileSize;
	if (fread(data, size, 1, mp->fp) < 1
	    || OutputTube_writeFully(out, data, size) < 0) {
	    return -1;
	}
	fileSize -= size;
    }
    return mp->bodySegment.length;
}
